! hde5.f90
! Explicit FDM transient solver for 3D tetrahedral mesh.
! Miroslav Broz (miroslav.broz@email.cz), Aug 2nd 2007

program hde5

use const
use types

! functions and some subroutines, which need modules due to dynamic allocation of arrays
use adjacent_faces_module
use depth_elems_module
use ftcs_step4_module
use insolation_module
use insolation_node_module
use inside_tetrahedron_module
use interp_cartesian_module
use interp_elems_module
use interp_tetrahedron_module
use normalize_module
use normals_faces_module
use normals_nodes_module
use read_node_module
use read_face_module
use read_elem_module
use rotate_around_axis_module
use surface_interior_module

implicit none

!=======================================================================
!
! variables
!

! input parameters
double precision :: rho,C,K_bulk,A_Bond,epsil,P,r,mesh_scale
double precision :: D,xi,f,delta,T_eq,T_star,dx,dy,dz,dh,omega,S_r
double precision :: t,tmin,tmax,dt,dtout,tout
double precision :: n_Sun_t0(3), n_rotate(3)
character(len=80) :: f_node,f_elem,f_face

! grid data
integer :: nnodes,nelems,nfaces
double precision, dimension(:,:), pointer :: nodes
integer, dimension(:,:), pointer :: elems, faces

integer, dimension(:), pointer :: surface_idx, interior_idx, adjacent_cnt
integer, dimension(:,:), pointer :: adjacent_idx
double precision, dimension(:,:), allocatable :: faces_normals, nodes_normals
double precision, dimension(:), allocatable :: areas
integer, dimension(:), allocatable :: depth_idx
integer, dimension(:,:), allocatable :: interp_idx

! internal variables
double precision, dimension(:), allocatable :: Temp, Temp_new, E, E_node
double precision, dimension(4,3) :: vertices
double precision, dimension(4) :: values
double precision, dimension(3) :: x
complex, dimension(5) :: a
complex :: z
double precision, dimension(:,:), allocatable :: Temp_interp
double precision :: T_1, n_Sun(3)
integer :: i,j,k,l, ierr
character(len=80) :: s
character(len=8) :: date
character(len=10) :: time
integer :: clock1, clock2

!=======================================================================
!
! print header
!

write(*,'(a)') '# hde5: Explicit FDM transient solver for 3D tetrahedral mesh.'
write(*,'(a,a)') '# version [] = ', 'Aug 2nd 2007'

call date_and_time(date, time)
call system_clock(clock1)
write(*,'(a,a,a,a)') '# start_time [YYYYMMDD HHMMSS.SSS] = ', date, ' ', time

!=======================================================================
!
! read input parameters
!
write(*,'(a)',advance='no') '# rho [kg/m^3] = '

! skip comments
do
  read(*,*,err=900,end=900) s
  if (s(1:1).ne.'#') exit
enddo

read(s,*,err=900,end=900) rho
write(*,*) rho

write(*,'(a)',advance='no') '# C [J/kg/K] = '
read(*,*,err=900,end=900) C
write(*,*) C

write(*,'(a)',advance='no') '# K [W/m/K] = '
read(*,*,err=900,end=900) K_bulk
write(*,*) K_bulk

write(*,'(a)',advance='no') '# A_Bond [] = '
read(*,*,err=900,end=900) A_Bond
write(*,*) A_Bond

write(*,'(a)',advance='no') '# epsilon [] = '
read(*,*,err=900,end=900) epsil
write(*,*) epsil

write(*,'(a)',advance='no') '# P [s] = '
read(*,*,err=900,end=900) P
write(*,*) P
20 format(a,g12.5)
write(*,20) '# P [h] = ', P/3600.d0

write(*,'(a)',advance='no') '# r [AU] = '
read(*,*,err=900,end=900) r
write(*,*) r

write(*,'(a)',advance='no') '# n_Sun_t0 [x,y,z] = '
read(*,*,err=900,end=900) n_Sun_t0(1),n_Sun_t0(2),n_Sun_t0(3)
call normalize(n_Sun_t0)
write(*,*) n_Sun_t0(1),n_Sun_t0(2),n_Sun_t0(3)

write(*,'(a)',advance='no') '# n_rotate [x,y,z] = '
read(*,*,err=900,end=900) n_rotate(1),n_rotate(2),n_rotate(3)
call normalize(n_rotate)
write(*,*) n_rotate(1),n_rotate(2),n_rotate(3)

write(*,'(a)',advance='no') '# dx [m] = '
read(*,*,err=900,end=900) dx
write(*,*) dx

write(*,'(a)',advance='no') '# dt [s] = '
read(*,*,err=900,end=900) dt
write(*,*) dt

write(*,'(a)',advance='no') '# dtout [s] = '
read(*,*,err=900,end=900) dtout
write(*,*) dtout

write(*,'(a)',advance='no') '# tmin [s] = '
read(*,*,err=900,end=900) tmin
write(*,*) tmin

write(*,'(a)',advance='no') '# tmax [s] = '
read(*,*,err=900,end=900) tmax
write(*,*) tmax
write(*,20) '# tmax [h] = ', tmax/3600.d0

write(*,'(a)',advance='no') '# mesh_scale [m] = '
read(*,*,err=900,end=900) mesh_scale
write(*,*) mesh_scale

write(*,'(a)',advance='no') '# file_with_nodes [*.node] = '
read(*,'(a)',err=900,end=900) f_node
write(*,*) f_node

write(*,'(a)',advance='no') '# file_with_elements [*.ele] = '
read(*,'(a)',err=900,end=900) f_elem
write(*,*) f_elem

write(*,'(a)',advance='no') '# file_with_surface_facets [*.face] = '
read(*,'(a)',err=900,end=900) f_face
write(*,*) f_face

!=======================================================================
!
! pre-calculate some parameters
!

! diffusivity
D = K_bulk/(rho*C)
write(*,20) '# D [m^2/s] = ', D

! von Neumann stability criterion
xi = 2*D*dt/(dx**2)
write(*,20) '# xi [] = ', xi

if (xi.gt.1) then
  write(*,'(a)') '# Warning: von Neumann stability criterion is NOT fulfilled!'
!  stop
endif

! equilibrium temperature
! pi R^2 (1-A) L_Sun / (4 pi r^2) = 4 pi R^2 epsilon sigma T_eq^4
T_eq = ((1-A_Bond)*S0 / (r**2 * 4*epsil*sigma))**0.25d0
write(*,20) '# T_eq [K] = ', T_eq

! subsolar temperature
T_star = sqrt(2.d0)*T_eq
write(*,20) '# T_star [K] = ', T_star

! diurnal/seasonal frequency
f = 1.d0/P
omega = 2.d0*pi*f
write(*,20) '# f [Hz] = ', f
write(*,20) '# f [rev/day] = ', day/P
write(*,20) '# omega [rad/s] = ', omega

! solar constant and flux (@ distance r)
S_r = S0/r**2
write(*,20) '# S0 [W/m^2] = ', S0
write(*,20) '# S_r [W/m^2] = ', S_r

! penetration depth
delta = sqrt(D/(pi*f))
write(*,20) '# delta [m] = ', delta

dy = dx
dz = dx
dh = dx
write(*,20) '# dy [m] = ', dy
write(*,20) '# dz [m] = ', dz
write(*,20) '# dh [m] = ', dh

!=======================================================================
!
! read grid data files
!
nullify(nodes)
nullify(faces)
nullify(elems)
call read_node(f_node,nodes,nnodes)
call read_face(f_face,faces,nfaces)
call read_elem(f_elem,elems,nelems)

25 format(a,i8)
write(*,25) '# number_of_nodes = ', nnodes
write(*,25) '# number_of_faces = ', nfaces
write(*,25) '# number_of_elements = ', nelems

! scale the nodes coordinates by mesh_scale factor
nodes = mesh_scale * nodes

! print size of the body
write(*,20) '# body_size_x [m] = ', maxval(nodes(:,1)) - minval(nodes(:,1))
write(*,20) '# body_size_y [m] = ', maxval(nodes(:,2)) - minval(nodes(:,2))
write(*,20) '# body_size_z [m] = ', maxval(nodes(:,3)) - minval(nodes(:,3))

! resolve surface/interior nodes
nullify(surface_idx)
nullify(interior_idx)
call surface_interior(faces, nnodes, surface_idx, interior_idx)

write(*,25) '# number_of_surface_nodes = ', size(surface_idx)
write(*,25) '# number_of_interior_nodes = ', size(interior_idx)

! write indices to stdout
do i = 1, size(surface_idx)
  write(*,'(a,i8,a,1x,i8)') '# surface_idx[', i, '] = ', surface_idx(i)
enddo

do i = 1, size(interior_idx)
  write(*,'(a,i8,a,1x,i8)') '# interior_idx[', i, '] = ', interior_idx(i)
enddo

! calculate NORMALS for all surface FACETS
allocate(faces_normals(nfaces,3))
allocate(areas(nfaces))
call normals_faces(nodes, faces, faces_normals, areas)

! write normals to stdout
35 format(a,i8,a,3(1x,g12.6))
do i = 1, nfaces
  write(*,35) '# faces_normals[', i, '] = ', faces_normals(i,:)
enddo

! list of ADJACENT FACETS for every surface node
nullify(adjacent_cnt)
nullify(adjacent_idx)
call adjacent_faces(faces, surface_idx, adjacent_cnt, adjacent_idx)

! write adjacent indices to stdout
write(*,'(a,i8)') '# max_adjacent_cnt = ', maxval(adjacent_cnt)
do i = 1, size(adjacent_idx(:,1))
  write(*,'(a,i8,a,20(1x,i8))') '# adjacent_idx[', i, '] = ', adjacent_idx(i,:)
enddo

! calculate NORMALS for all surface NODES (as an average of adjacent facets)
allocate(nodes_normals(size(surface_idx), 3))
call normals_nodes(faces_normals, adjacent_cnt, adjacent_idx, nodes_normals)

! write normals to stdout
do i = 1, size(nodes_normals(:,1))
  write(*,35) '# nodes_normals[', i, '] = ', nodes_normals(i,:)
enddo

! list of ELEMENTS in the DEPTH below surface nodes
allocate(depth_idx(size(surface_idx)))
call depth_elems(nodes, elems, surface_idx, nodes_normals, dh, depth_idx)

! write interpolation points to stdout
do i = 1, size(surface_idx)
  write(*,35) '# depth_x[', i, '] = ', nodes(surface_idx(i),:) - dh * nodes_normals(i,:)
enddo

! write elements indices to stdout
do i = 1, size(depth_idx)
  write(*,'(a,i8,a,1x,i8)') '# depth_idx[', i, '] = ', depth_idx(i)
enddo

! list of ELEMENTS for temperature INTERPOLATION
allocate(interp_idx(size(interior_idx), 6))
call interp_elems(nodes, elems, interior_idx, dx, dy, dz, interp_idx)

! write interpolation points to stdout
do i = 1, size(interior_idx)
  write(*,35) '# interp_x_plus_dx[' , i, '] = ', nodes(interior_idx(i),:) + (/+dx, 0.d0, 0.d0/)
  write(*,35) '# interp_x_minus_dx[', i, '] = ', nodes(interior_idx(i),:) + (/-dx, 0.d0, 0.d0/)
  write(*,35) '# interp_x_plus_dy[' , i, '] = ', nodes(interior_idx(i),:) + (/0.d0, +dy, 0.d0/)
  write(*,35) '# interp_x_minus_dy[', i, '] = ', nodes(interior_idx(i),:) + (/0.d0, -dy, 0.d0/)
  write(*,35) '# interp_x_plus_dz[' , i, '] = ', nodes(interior_idx(i),:) + (/0.d0, 0.d0, +dz/)
  write(*,35) '# interp_x_minus_dz[', i, '] = ', nodes(interior_idx(i),:) + (/0.d0, 0.d0, -dz/)
enddo

! write elements to stdout
do i = 1, size(interp_idx(:,1))
  write(*,'(a,i8,a,6(1x,i8))') '# interp_idx[', i, '] = ', interp_idx(i,:)
enddo

! indicate CPU_time consumed so far
call system_clock(clock2)
write(*,20) '# CPU_time [s] = ', 0.001d0*(clock2-clock1)

!=======================================================================
!
! INITIAL CONDITION (@ t=0)
!

! allocate temperature arrays
allocate(Temp(nnodes))
allocate(Temp_new(nnodes))

t = tmin
Temp = T_eq

call rotate_around_axis(n_rotate, omega*t, n_Sun_t0, n_Sun)

!
! an initial dump
!
write(*,'(a)') '# t [s] & node_id [] & T [K]'
write(*,20) '# t [s] = ', t
55 format (a,3(1x,g14.8))
write(*,55) '# n_Sun [x,y,z] = ', n_Sun
50 format(g14.8,1x,i8,1x,g14.8)
do j = 1, nnodes
  write(*,50) t, j, Temp(j)
enddo
write(*,*)

tout = t + dtout

!=======================================================================
!
! MAIN LOOP
!

! allocate arrays for insolation and temperature interpolation 
allocate(E(nfaces))
allocate(E_node(size(surface_idx)))
allocate(Temp_interp(size(interior_idx),6))

main_loop: do while (t.le.(tmax - 0.5d0*dt))
  t = t + dt

!=======================================================================
!
! BOUNDARY CONDITION (for surface nodes)
!

! calculate insolation (for all surface FACETS)
! K (dT/dr)_surface + epsilon sigma T^4 = (1-A) * E(t) * n_normal(r)

  call rotate_around_axis(n_rotate, omega*t, n_Sun_t0, n_Sun)

  all_faces: do j = 1, nfaces
    E(j) = insolation(S_r, n_Sun, faces_normals(j,:))
  enddo all_faces

  surface_nodes: do j = 1, size(surface_idx)   ! surface nodes index
    i = surface_idx(j)                         ! all nodes index

! average node insolation (contributions from all adjacent facets)
    E_node(j) = insolation_node(E, areas, adjacent_cnt(j), adjacent_idx(j,:))

! interpolate the temperature T_1 in the depth (just below the node, @ coords x)
! the PREVIOUS surface temperature might be also used in this interpolation
    x = nodes(i,:) - dh * nodes_normals(j,:)    ! interpolation point
    do k = 1, 4
      l = elems(depth_idx(j),k)                 ! which nodes/temperatures?
      vertices(k,:) = nodes(l,:)
      values(k) = Temp(l)
    enddo
    T_1 = interp_tetrahedron(vertices, values, x)

! K (dT/dr)_surface + epsilon sigma T^4 = (1-A) * E(t) * n_perp(r)
! we express the derivative dT/dr as a difference (T_0-T_1)/dh
! for K<>0 we have to solve a 4-th order polynomial (by Laguerre method)
! see Capek (2007) for explanation
    a(1) = cmplx( - (1.d0-A_Bond)/(epsil*sigma) * E_node(j) - K_bulk/(epsil*sigma) * 1.d0/dh * T_1 )
    a(2) = cmplx( + K_bulk/(epsil*sigma) * 1.d0/dh )
    a(3) = (0.,0.)
    a(4) = (0.,0.)
    a(5) = (1.,0.)
! take the subsolar temperature as a 1st guess
    T_star = ((1.d0-A_Bond)/(epsil*sigma) * E_node(j))**0.25d0
    z = cmplx(T_star)

    call laguer(a,4,z,k)

    Temp(i) = dble(z)

  enddo surface_nodes

!=======================================================================
!
! TIME STEP (for interior nodes)
!

! interpolate the temperature
  call interp_cartesian(Temp, nodes, elems, interior_idx, interp_idx, dx, dy, dz, Temp_interp)

! explicit time step
  call ftcs_step4(Temp, Temp_interp, dt, dx, D, surface_idx, interior_idx, Temp_new)

! move the whole array
  Temp = Temp_new

!
! print output every dtout (for ALL nodes)
!
  if (t.ge.(tout-0.5d0*dt)) then
    write(*,20) '# t [s] = ', t
    write(*,20) '# percent_done [%] = ', 100.d0*(t-tmin)/(tmax-tmin)
    write(*,55) '# n_Sun [x,y,z] = ', n_Sun
65  format(a,i8,a,g14.8)
    do j = 1, nfaces
      write(*,65) '# E[', j, '] = ', E(j)
    enddo
    do j = 1, size(surface_idx)
      write(*,65) '# E_node[', j, '] = ', E_node(j)
    enddo
    do j = 1, nnodes
      write(*,50) t, j, Temp(j)
    enddo
    write(*,*)
    tout = tout + dtout
  endif

enddo main_loop

!=======================================================================
!
! print footer
!

call date_and_time(date, time)
call system_clock(clock2)

write(*,20) '# CPU_time [s] = ', 0.001d0*(clock2-clock1)
write(*,'(a,a,a,a)') '# stop_time [YYYYMMDD HHMMSS.SSS] = ', date, ' ', time

stop

!=======================================================================
!
! error handlers
!

900 write(*,'(a)') '# Error: reading input parameters.'
stop


end program hde5


